/*
    Copyright (C) 2002 Terratron Technologies Inc.  All Rights Reserved.
    Author: Steven Osman
 
    This file is part of sps2.

    Before using this file you MUST agree to the license agreement in
    the file LICENSE provided with this package.

    For more information visit:
    http://www.playstation2-linux.com/projects/sps2/
    http://window.terratron.com/~sosman/ps2linux/
    http://www.terratron.com/
*/
#define EXPORT_SYMTAB

#include <linux/module.h>

#include "sps2_mod.h"

static int sauce_file_mmap(struct file *pFile,
			   struct vm_area_struct *pVMArea);
static int sauce_file_open(struct inode *pInode, 
			   struct file *pFile);
static int sauce_file_release(struct inode *pInode, 
			      struct file *pFile);
static int sauce_file_ioctl(struct inode *pInode, struct file *pFile,
			    unsigned int uiCmd, unsigned long ulArg);
static unsigned long sauce_vm_nopage(struct vm_area_struct *pVMArea,
				     unsigned long ulAddress,
				     int iWriteAccess);
memoryMap_t *allocateMap(unsigned long ulStart, int iBlockCount,
			 int iBlockOrder,
			 physicalMemoryMap_t *pPhysicalMemoryMap);
void freeMap(memoryMap_t *pMap, physicalMemoryMap_t *pPhysicalMemoryMap);
void adjustPhysicalMemoryMap(physicalMemoryMap_t *pPhysicalMemoryMap,
			     unsigned long ulStart,
			     int iNPages, int iFree);
int init_module(void);
void cleanup_module(void);



#define SPS2_MOD_VERSION_MAJOR 0
#define SPS2_MOD_VERSION_MINOR 2
#define SPS2_MOD_VERSION_REVISION 0

#define DEVICE_NAME "sauce_char"

//EXPORT_NO_SYMBOLS;

/**
 * This function is used to remove printks when debugging
 * is disabled
 */
void null_printk(const char *pcFormat, ...) {
}

#define DEBUG_OUTPUT null_printk
// #define DEBUG_OUTPUT printk
/**
 * The operations possible on our file.
 */
static struct file_operations fileOperations={
  ioctl:	sauce_file_ioctl,
  open:		sauce_file_open,
  release:	sauce_file_release,
  mmap:         sauce_file_mmap,
};

static struct vm_operations_struct vmOperations={
  nopage:	sauce_vm_nopage,
};

/**
 * The major number of our device
 */
static int iDeviceMajorNumber;


/**
 * Pointers to the DMA Channel Control Registers
 */
static volatile unsigned int *apuiDMACHCR[]={
  (unsigned int *) KSEG1ADDR(0x10008000),
  (unsigned int *) KSEG1ADDR(0x10009000),
  (unsigned int *) KSEG1ADDR(0x1000a000),
  (unsigned int *) KSEG1ADDR(0x1000b000),
  (unsigned int *) KSEG1ADDR(0x1000b400),
  (unsigned int *) KSEG1ADDR(0x1000c000),
  (unsigned int *) KSEG1ADDR(0x1000c400),
  (unsigned int *) KSEG1ADDR(0x1000c800),
  (unsigned int *) KSEG1ADDR(0x1000d000),
  (unsigned int *) KSEG1ADDR(0x1000d400),
};

int (*sps2ValidateDMA)(int iChannel, sps2uint32 ui32CHCR, int iSizeLimit,
		   physicalMemoryMap_t *pPhysicalMemoryMap,
		   sps2ValidateDMAInfo *pRequest);
EXPORT_SYMBOL(sps2ValidateDMA);

/**
 * Allocates a chunk of memory.
 * On success, it returns a memoryMap_t that pointer that must
 * be freed with freeMap.  On failure, it returns null.
 * ulStart is the starting address in the user space program
 * iBlockCount is the number of blocks to allocate
 * iBlockOrder is the order of the blocks to allocate.
 *             0 = 2^0 pages (i.e. 1 page).  5=2^5 pages = 32 pages
 */
memoryMap_t *allocateMap(unsigned long ulStart, int iBlockCount,
			 int iBlockOrder,
			 physicalMemoryMap_t *pPhysicalMemoryMap) {
  memoryMap_t *pMap=vmalloc(sizeof(memoryMap_t)+
			    sizeof(unsigned long)*(iBlockCount-1));
  int iPageLooper;
  unsigned long *pulBlocks;

  printk("sps2: Allocating a %i order-%i block list\n",iBlockCount,
	 iBlockOrder);

  if (pMap) {
    pMap->ulStart=ulStart;
    pMap->iBlockCount=iBlockCount;
    pMap->iBlockSize=1 << (PAGE_SHIFT+iBlockOrder);
    pMap->iBlockOrder=iBlockOrder;
    pulBlocks=&(pMap->aulBlocks);
    pMap->piReferenceCounter=vmalloc(sizeof(int));

    if (!pMap->piReferenceCounter) {
      vfree(pMap);
      pMap=0;
    }

    *(pMap->piReferenceCounter)=0;
  }

  if (pMap) {
    for (iPageLooper=0;iPageLooper<iBlockCount;iPageLooper++) {
      pulBlocks[iPageLooper]=__get_free_pages(GFP_KERNEL, iBlockOrder);
      
      if (!pulBlocks[iPageLooper]) {
	for (--iPageLooper;iPageLooper>=0;iPageLooper--) {
	  free_pages(pulBlocks[iPageLooper], iBlockOrder);
	}

	vfree(pMap);
	pMap=0;
	break;
      }

      adjustPhysicalMemoryMap(pPhysicalMemoryMap,
			      virt_to_bus((void *) pulBlocks[iPageLooper]) ,
			      1 << iBlockOrder, 0);
    }
  }
  
  return pMap;
}

/**
 * Releases the memory and destroys the memoryMap_t created by
 * allocateMap.
 * pMap is the map to release.
 */
void freeMap(memoryMap_t *pMap, physicalMemoryMap_t *pPhysicalMemoryMap) {
  if (pMap->mapType==MAP_TYPE_MEMORY) {
    int iPageLooper;
    unsigned long *pulBlocks=&(pMap->aulBlocks);

    if (--(*(pMap->piReferenceCounter))==0) {
      printk("sps2: Clearing a %i %i-block list\n",pMap->iBlockCount,
	     pMap->iBlockSize);

      vfree(pMap->piReferenceCounter);
      
      for (iPageLooper=0;iPageLooper<pMap->iBlockCount;iPageLooper++) {
	free_pages(pulBlocks[iPageLooper], pMap->iBlockOrder);

	adjustPhysicalMemoryMap(pPhysicalMemoryMap, 
				virt_to_bus((void *) pulBlocks[iPageLooper]),
				1 << pMap->iBlockOrder, 1);
      }
    }
  }

  vfree(pMap);
}

/**
 * This method searches through the other open files for this process
 * to see if it can find a physical memory map. If not, then it creates
 * a new one.  The reference counter for the map is incremented.
 * pFile is the current open file.
 */
physicalMemoryMap_t *initPhysicalMemoryMap(struct file *pFile) {
  int iLooper;
  physicalMemoryMap_t *pPhysicalMemoryMap=0;

  for (iLooper=0;iLooper<current->files->max_fds;iLooper++) {
    struct file *pSearchFile=current->files->fd[iLooper];
    
    if (pSearchFile && (pSearchFile->f_dentry==pFile->f_dentry)) {
      fileData_t *pFileData=(fileData_t *) pSearchFile->private_data;
      if (pFileData &&
	  pFileData->pPhysicalMemoryMap) {
	pPhysicalMemoryMap=pFileData->pPhysicalMemoryMap;
      }
    }
  }

  if (!pPhysicalMemoryMap) {
    pPhysicalMemoryMap=vmalloc(sizeof(physicalMemoryMap_t));
    memset(pPhysicalMemoryMap,0,sizeof(physicalMemoryMap_t));
  }

  pPhysicalMemoryMap->iReferenceCounter++;

  return pPhysicalMemoryMap;
}

/**
 * Adds or removes entries from the physical memory map when data is
 * allocated or freed for this process.
 * pPhysicalMemoryMap is the map to update
 * ulStart is the start address of the area.
 * iNPages is the number of pages in the area
 * iFree is non-zero when the memory is being freed, zero when it is
 *  being allocated
 */
void adjustPhysicalMemoryMap(physicalMemoryMap_t *pPhysicalMemoryMap,
			     unsigned long ulStart,
			     int iNPages, int iFree) {
  int iLooper;
  int iPage;
  unsigned char *pucByte;
  int iShift;

  iPage=ulStart>>12;	   

  for (iLooper=0;iLooper<iNPages;iLooper++) {
    if (iPage >= 8192) {
      printk("sps2:adjustPhysicalMemoryMap Invalid pointer %08lx count %i\n",
	     ulStart, iNPages);
      return;
    }

    pucByte=pPhysicalMemoryMap->aucMap+(iPage >> 2);
    iShift=2 * (iPage & 3);

    if (iFree) {
      *pucByte &= ~(MEMORY_MASK << iShift);
    } else if (iLooper) {
      *pucByte |= (MEMORY_CONTINUE << iShift);
    } else {
      *pucByte |= (MEMORY_START << iShift);
    }

    iPage++;
  }    
}

// #include "dmachecker.c"

/**
 * Called when a page has faulted.
 */
static unsigned long sauce_vm_nopage(struct vm_area_struct *pVMArea,
				     unsigned long ulAddress,
				     int iWriteAccess) {
  fileData_t *pFileData=(fileData_t *) pVMArea->vm_file->private_data;
  unsigned long ulOffset=ulAddress-pVMArea->vm_start;
  unsigned long ulReturn=0;

  DEBUG_OUTPUT("sps2:Page fault:\nAddress: %08lx\nOffset: %08lx\n",
	 ulAddress, pVMArea->vm_offset);


  switch (pVMArea->vm_offset & (~0xff)) {
  case SPS2_MAP_MEMORY:
    {
      struct list_head *pSearch;
      memoryMap_t *pMap=0;
      int iBlockNumber;
      
      list_for_each(pSearch, &pFileData->mapList) {
	if (list_entry(pSearch, memoryMap_t, list)->ulStart==
	    pVMArea->vm_start) {
	  pMap=list_entry(pSearch, memoryMap_t, list);
	  break;
	}
      }

      if (pMap) {
	iBlockNumber=ulOffset / pMap->iBlockSize;
      
	ulReturn=((&pMap->aulBlocks)[iBlockNumber]) +
	  (ulOffset % pMap->iBlockSize);
      }
    }
  }

  DEBUG_OUTPUT("sps2:Return: %08lx\n",ulReturn);

  return ulReturn;
}

/**
 * Called when a program attempts to mmap the device file to memory.
 * We have different kinds of mapping:  One which allocates memory, the
 * others which request access to some of the register memory areas.
 */
static int sauce_file_mmap(struct file *pFile,
		      struct vm_area_struct *pVMArea) { 
  fileData_t *pFileData=(fileData_t *) pFile->private_data;
  memoryMap_t *pMap=0;
  unsigned long ulSize=pVMArea->vm_end-pVMArea->vm_start;

  // If we don't have a version number for compatibility yet, we
  // fail
  if (pFileData->iVersionMajor<0 || pFileData->iVersionMinor<0) {
    printk("sps2:mmap Rejecting command because client %s with pid %i\n"
	   "did not request a version compatibility.\n",
	   current->comm ? current->comm : "(unknown)", (int) current->pid);
    return -EINVAL;
  }

  switch (pVMArea->vm_offset & (~0xff)) {
  case SPS2_MAP_MEMORY:
    {
      //      int iBlockOrder = pVMArea->vm_offset & 0x7;
      int iBlockOrder = 0;
      int iBlockSize = 1 << (PAGE_SHIFT+iBlockOrder);

      if (ulSize % iBlockSize) {
	return -EINVAL;
      }

      pMap=allocateMap(pVMArea->vm_start,
		       ulSize / iBlockSize,
		       iBlockOrder, pFileData->pPhysicalMemoryMap);
      pMap->mapType=MAP_TYPE_MEMORY;
    }
    break;

  case SPS2_MAP_EE_REGISTERS:
    if (ulSize==SPS2_MAP_EE_REGISTERS_SIZE) {
      remap_page_range(pVMArea->vm_start, SPS2_MAP_EE_REGISTERS,
		       SPS2_MAP_EE_REGISTERS_SIZE, PAGE_USERIO);
      printk("Mapping physical %p to virtual %p\n",
	     (void *) SPS2_MAP_EE_REGISTERS, (void *) pVMArea->vm_start);
      return 0;
    } else {
      return -EINVAL;
    }
    break;

  case SPS2_MAP_VU_MEMORY:
    if (ulSize==SPS2_MAP_VU_MEMORY_SIZE) {
      remap_page_range(pVMArea->vm_start, SPS2_MAP_VU_MEMORY,
		       SPS2_MAP_VU_MEMORY_SIZE, PAGE_USERIO);
      return 0;
    } else {
      return -EINVAL;
    }
    break;

  case SPS2_MAP_GS_REGISTERS:
    if (ulSize==SPS2_MAP_GS_REGISTERS_SIZE) {
      remap_page_range(pVMArea->vm_start, SPS2_MAP_GS_REGISTERS,
		       SPS2_MAP_GS_REGISTERS_SIZE, PAGE_USERIO);

      return 0;
    } else {
      return -EINVAL;
    }
    break;

  case SPS2_MAP_SCRATCH_PAD:
    if (ulSize==2*SPS2_MAP_SCRATCH_PAD_SIZE) {
      // Some of this code borrowed from ps2dev.c

      if (pVMArea->vm_start % 16384) {
	pVMArea->vm_start-=pVMArea->vm_start % 16384;
	pVMArea->vm_start+=16384;
      }

      pVMArea->vm_end=pVMArea->vm_start+SPS2_MAP_SCRATCH_PAD_SIZE;
      pVMArea->vm_flags |= VM_IO;
      pgprot_val(pVMArea->vm_page_prot) = (pgprot_val(pVMArea->vm_page_prot) &
					   ~_CACHE_MASK) | _CACHE_UNCACHED;

      remap_page_range(pVMArea->vm_start, SPS2_MAP_SCRATCH_PAD,
		       SPS2_MAP_SCRATCH_PAD_SIZE, pVMArea->vm_page_prot);

      return 0;
    } else {
      return -EINVAL;
    }
    break;

  default:
    {
      // The user may be asking to remap an area, for example, in order to change
      // the area to uncached.  We need to make sure that we find the exact area,
      // and that it's the same size.  We'll look for it by searching by base physical
      // address.
      unsigned long ulBasePhysical=
	virt_to_bus((void *) ((unsigned long) pVMArea->vm_offset & (~0xff)));
      memoryMap_t *pSearchMap=0;
      struct list_head *pSearch;

      list_for_each(pSearch, &pFileData->mapList) {
	if (virt_to_bus((void *) list_entry(pSearch, memoryMap_t,
					    list)->aulBlocks)==
	    ulBasePhysical) {
	  pSearchMap=list_entry(pSearch, memoryMap_t, list);
	  break;
	}
      }

      if (!pSearchMap) {
	printk("sps2: Remap with base bus address of %08lx not found\n",
	       ulBasePhysical);
	return -EINVAL;
      }

      // Memory area size must match
      if ((pVMArea->vm_end-pVMArea->vm_start) !=
	  (pSearchMap->iBlockSize*pSearchMap->iBlockCount)) {
	printk("sps2: Remap area found, but wrong size."
	       "  Got size %08lx but expected size %08x\n",
	       (pVMArea->vm_end-pVMArea->vm_start),
	       (pSearchMap->iBlockSize*pSearchMap->iBlockCount));
	return -EINVAL;
      }

      // Block order must match
      if (pSearchMap->iBlockOrder!=(pVMArea->vm_offset & 0x7)) {
	printk("sps2: Remap area found, but wrong block order."
	       "  Got %i but expected %i\n",
	       (int) (pVMArea->vm_offset & 0x7), pSearchMap->iBlockOrder);
	return -EINVAL;
      }


      // For memory maps, offset should not contain the address.
      // Strip everything other than flags
      pVMArea->vm_offset &= 0xff;

      printk("sps2: Found and remapping memory area\n");

      // We found the map, everything looks in order.  Now we just want to
      // clone it and take the bits that are specific to this remap.
      pMap=vmalloc(sizeof(memoryMap_t)+sizeof(unsigned long)*
		   (pSearchMap->iBlockCount-1));

      if (pMap) {
	memcpy(pMap, pSearchMap, sizeof(memoryMap_t)+
	       sizeof(unsigned long)*(pSearchMap->iBlockCount-1));
      }
    }
    break;
  }

  if (!pMap) {
    return -ENOMEM;
  }

  if (pVMArea->vm_offset & SPS2_MAP_UNCACHED) {
    pgprot_val(pVMArea->vm_page_prot) =
      (pgprot_val(pVMArea->vm_page_prot) & ~_CACHE_MASK) | _CACHE_UNCACHED;
  } else {
    pgprot_val(pVMArea->vm_page_prot) =
      (pgprot_val(pVMArea->vm_page_prot) & ~_CACHE_MASK) |
      _CACHE_CACHABLE_NONCOHERENT;
  }

  // Avoid swapping out of this memory
  pVMArea->vm_flags|=VM_LOCKED;

  pFileData->iMapCount++;
  list_add(&pMap->list, &pFileData->mapList);
  pVMArea->vm_ops=&vmOperations;
  pVMArea->vm_file=pFile;
  pMap->ulStart=pVMArea->vm_start;
  (*(pMap->piReferenceCounter))++;

  pFile->f_count++;

  printk("sps2: Map returning memory beginning at virtual %08lx\n",
	 pVMArea->vm_start);

  return 0;
}

/**
 * Called when the device file is opened.  Initializes the private_data
 * field.
 */
static int sauce_file_open(struct inode *pInode, 
		     struct file *pFile) {
  fileData_t *pFileData=vmalloc(sizeof(fileData_t));

  if (!pFileData) {
    return -ENOMEM;
  }

  pFileData->iVersionMinor=pFileData->iVersionMajor=-1;

  INIT_LIST_HEAD(&pFileData->mapList);
  pFileData->iMapCount=0;
  pFileData->pPhysicalMemoryMap=initPhysicalMemoryMap(pFile);

  pFile->private_data=pFileData;

  MOD_INC_USE_COUNT;

  printk("sps2:open Opened by client %s with pid %i\n",
	 current->comm ? current->comm : "(unknown)", (int) current->pid);
  return 0;
}


/**
 * This is called when a file is being closed.
 * Make sure to release any memory associated with this
 * file.
 */
static int sauce_file_release(struct inode *pInode, 
			 struct file *pFile) {
  fileData_t *pFileData=(fileData_t *) pFile->private_data;
  struct list_head *pLoop, *pTemp;

  MOD_DEC_USE_COUNT;

  list_for_each_safe(pLoop, pTemp, &pFileData->mapList) {
    list_del(pLoop);

    freeMap(list_entry(pLoop, memoryMap_t, list),
	    pFileData->pPhysicalMemoryMap);
  }

  if (!(--(pFileData->pPhysicalMemoryMap->iReferenceCounter))) {
    vfree(pFileData->pPhysicalMemoryMap);
  }

  vfree(pFileData);

  printk("sps2:release Closed by client %s with pid %i\n",
	 current->comm ? current->comm : "(unknown)", (int) current->pid);
  return 0;
}

/**
 * Performs an ioctl call on our device file.
 */
static int sauce_file_ioctl(struct inode * pInode, struct file *pFile,
			    unsigned int uiCommand,
			    unsigned long ulArgument) {
  fileData_t *pFileData=(fileData_t *) pFile->private_data;

  // If we don't have a version number for compatibility yet, we
  // only allow the IOCTL command to require version compatibility
  if ((pFileData->iVersionMajor<0 || pFileData->iVersionMinor<0) &&
      (uiCommand!=SPS2_IOCTL_QUERY_SUPPORTED_VERSION)) {
    printk("sps2:ioctl Rejecting command because client %s with pid %i\n"
	   "did not request a version compatibility.\n",
	   current->comm ? current->comm : "(unknown)", (int) current->pid);
    return -EINVAL;
  }

  switch (uiCommand) {
  case SPS2_IOCTL_PHYSICAL_ADDRESS:
    {
      struct list_head *pSearch;
      memoryMap_t *pMap=0;
      sps2PhysicalAddressInfo physicalRequest;
      int iLooper;

      if (copy_from_user(&physicalRequest, (void *) ulArgument,
			  sizeof(sps2PhysicalAddressInfo))) {
	printk("sps2:ioctl.PHYSICAL_ADDRESS: Invalid structure pointer "
	       "%08lx\n",
	       ulArgument);
	return -EINVAL;
      }

      list_for_each(pSearch, &pFileData->mapList) {
	if (list_entry(pSearch, memoryMap_t, list)->ulStart==
	    (unsigned long) physicalRequest.pvVirtualStart) {
	  pMap=list_entry(pSearch, memoryMap_t, list);
	  break;
	}
      }

      if (!pMap) {
	printk("sps2:ioctl.PHYSICAL_ADDRESS: Area %08lx not found.\n",
	       (unsigned long) physicalRequest.pvVirtualStart);
	return -EINVAL;
      }

      printk("sps2:Returning physical of %i blocks of %i bytes\n",
	     pMap->iBlockCount, pMap->iBlockSize);

      if (physicalRequest.iBufferSize!=
	  pMap->iBlockCount * sizeof(unsigned long)) {
	printk("sps2:ioctl.PHYSICAL_ADDRESS: Invalid buffer size.  "
	       "Got %i but expected %i\n", 
	       physicalRequest.iBufferSize,
	       pMap->iBlockCount*sizeof(unsigned long));
	return -EINVAL;
      }


      for (iLooper=0;iLooper<pMap->iBlockCount;iLooper++) {
	put_user(virt_to_bus((void *) (&pMap->aulBlocks)[iLooper]),
		 physicalRequest.pulPhysicalPageList+iLooper);
      }
    }
    break;

  case SPS2_IOCTL_FLUSH_CACHE:
    flush_cache_all();
    break;

  case SPS2_IOCTL_WAIT_FOR_DMA:
    {
      if (ulArgument>9) {
	return -EINVAL;
      }

      while ((*(apuiDMACHCR[ulArgument])) & 0x100) {
	current->policy |= SCHED_YIELD;
      	schedule();
      }

      return 0;
    }

  case SPS2_IOCTL_QUERY_SUPPORTED_VERSION:
    {
      int iLooper;

      sps2VersionInfo version;
      version.ulVersion=ulArgument;
      if (version.v.uiWillShare) {
	// See if we have another instance of the file that is
	// opened for the same version
	for (iLooper=0;iLooper<current->files->max_fds;iLooper++) {
	  struct file *pSearchFile=current->files->fd[iLooper];
	  
	  if (pSearchFile && (pSearchFile->f_dentry==pFile->f_dentry)) {
	    fileData_t *pFileData=(fileData_t *) pSearchFile->private_data;
	    if (pFileData &&
		pFileData->iVersionMajor==version.v.usMajor &&
		pFileData->iVersionMinor==version.v.usMinor &&
		pFileData->iShare) {
	      printk("sps2:ioctl:version Returning file descriptor %i "
		     "because\n it meets the requested version requirements\n",
		     iLooper);
	      return iLooper;
	    }
	  }
	}
      }

      // We haven't found an open instance of the driver for this version
      // Run the version check and return

      // The current module code only supports one version.
      if (version.v.usMajor!=SPS2_MOD_VERSION_MAJOR ||
	  version.v.usMinor!=SPS2_MOD_VERSION_MINOR) {
	printk("sps2:ioctl: ""Module version %i.%i.%i "
	       "cannot support requested version %i.%i\n",
	       SPS2_MOD_VERSION_MAJOR, SPS2_MOD_VERSION_MINOR, 
	       SPS2_MOD_VERSION_REVISION, version.v.usMajor,
	       version.v.usMinor);
	// ENOSYS = function not implemented, i.e. we don't do
	// this version.
	return -ENOSYS;
      }

      pFileData->iVersionMajor=version.v.usMajor;
      pFileData->iVersionMinor=version.v.usMinor;
      pFileData->iShare=version.v.uiWillShare;

      // We need to return the current file descriptor
      for (iLooper=0;iLooper<current->files->max_fds;iLooper++) {
	if (pFile==current->files->fd[iLooper]) {
	  return iLooper;
	}
      }
    }
    break;

  case SPS2_IOCTL_VALIDATE_DMA:
    {
      sps2ValidateDMAInfo info;
      int iReturnValue;

      if (!sps2ValidateDMA) {
	printk("sps2:ioctl.VALIDATE_DMA: A DMA Validation Module is not "
	       "loaded\n");
	return -ENOSYS;
      }

      if (copy_from_user(&info, (void *) ulArgument,
			  sizeof(sps2ValidateDMAInfo))) {
	printk("sps2:ioctl.VALIDATE_DMA: Invalid structure pointer %08lx\n",
	       ulArgument);
	return -EINVAL;
      }

      iReturnValue=sps2ValidateDMA(info.iChannel, info.ui32CHCR,
				   info.iSizeLimit,
				   pFileData->pPhysicalMemoryMap, &info);

      if (copy_to_user((void *) ulArgument, &info,
		       sizeof(sps2ValidateDMAInfo))) {
	printk("sps2:ioctl.VALIDATE_DMA: Couldn't copy return: "
	       "Invalid structure pointer %08lx\n",
	       ulArgument);
	return -EINVAL;
      }

      return iReturnValue;
    }

    break;


  default:
    printk("sps2:ioctl: Invalid command %i specified\n",
	   uiCommand);
    return -EINVAL;
  }

  return 0;
}

/**
 * Initializes the module.
 * Basically replaces the system call, opens up the shared memory area and
 * indicates that the module is loaded.
 */
int init_module(void) {
  /* Register the character device (atleast try) */
  iDeviceMajorNumber = register_chrdev(0, DEVICE_NAME, &fileOperations);

  /* Negative values signify an error */
  if (iDeviceMajorNumber < 0) {
    printk ("%s device failed with %d\n",
	    DEVICE_NAME, iDeviceMajorNumber);
    return iDeviceMajorNumber;
  }

  printk("sps2: Module version %i.%i.%i loaded as major device %i\n",
	 SPS2_MOD_VERSION_MAJOR, SPS2_MOD_VERSION_MINOR,
	 SPS2_MOD_VERSION_REVISION, iDeviceMajorNumber);

  sps2ValidateDMA=0;

  return 0;
}

/**
 * Removes the module.
 * Replaces the system call with the original system call and indicates
 * that the module is no longer loaded.
 */
void cleanup_module(void) {
  int iReturn;

  /* Unregister the device */
  iReturn = unregister_chrdev(iDeviceMajorNumber, DEVICE_NAME);
 
  if (iReturn<0) {
    printk("sps2: Error in unregister_chrdev of %s: %d\n", DEVICE_NAME, 
	   iReturn);
  }

  printk("sps2: Module version %i.%i.%i with major device %i unloaded\n",
	 SPS2_MOD_VERSION_MAJOR, SPS2_MOD_VERSION_MINOR,
	 SPS2_MOD_VERSION_REVISION, iDeviceMajorNumber);
}
